;      emm.asm   Revision 1.0  8/23/95
;
; This software was written by Symbios Logic Inc. to develop and test new
; products.  Symbios Logic assumes no liability for its use.  This software
; is released to the public domain to illustrate certain programming
; techniques for the 53c720 in target mode.
;
; Revision 1.0 8/23/95  Scott Rawlings
; Initial release.
;
;Prototypes
;
;int check_emm(void);
;	Checks to see if an EMS manager is functioning in the system.
;	0 indicates EMM is present, 1 indicates error.
;
;unsigned long int get_page_frame(void);
;	Function returns the base address segment of the first of four physical
;	windows into logical memory.  Page 0 is at offset 0, page 1 is at offset
;	16384, etc.
;
;unsigned long int check_free_pages(void);
;	Returns an unsigned long int that indicates the number of free
;	16 KB pages available in memory
;
;unsigned long int allocate_pages(unsigned int required_pages);
;	Takes the number of required pages passed to it and makes the request
;	to the EMS manager.  If sucessful, the EMM handle that identifies
;	this group of pages is returned as an unsigned long int.  If the call
;	fails, a zero is returned.
;
;int seek(unsigned long int l, int p, unsigned long int EMM_handle);
;	Using the EMM handle to reference the goup of pages, this function
; 	attempts to assign physical memory windows to point to logical pages
; 	in memory.  l gives the initial logical page to map, and p gives the
;	number of mappings to perform.  For example, if l = 5 and p = 2, then
;	logical page 5 would be mapped to physical page 0, and logical page
;	6 would be mapped to physical page 1.  A return value of 0 indicates
; 	success, a 1 indicates an error.
;
;int release(unsigned long int EMM_handle);
;	Releases control of the portion of memory referenced by the EMM handle.

dos			equ	21h	; interrupt which invokes DOS
dosexit			equ	4ch	; DOS function to exit program cleanly
EMM_interrupt		equ	67h	; interrupt to invoke EMM functions
get_page_segment	equ	41h	; get page frame segment address function
free_pages		equ	42h	; get free pages count function
get_handle		equ	43h	; allocate pages and get handle function
map_page		equ	44h	; map logical page to physical page
deallocate_pages	equ	45h	; deallocate pages function

	.model	small

	.code
	public	_check_emm
_check_emm	proc	near		; begin procedure check_emm
	push	bp
	push	si
	push	di

	push	ds
	push	cs
	pop	ds

	mov	ah, 35h			; issue get interrupt
	mov	al, EMM_interrupt	; vector
	int	21h

	mov	di,000ah		; use this segment in
					; es returned by DOS,
					; place the "device name
					; field" offset in di

	lea	si, ASCII_device_name	; place the offset of the
					; EMMXXXX0 name string in si,
					; the segment is already in ds

	mov	cx, 8			; compare the name strings
	cld
	repe	cmpsb
	jne	error_exit0

	xor	ax, ax			; EMM is present in the system
	pop 	ds
	pop	di
	pop	si
	pop	bp
	ret

error_exit0:

	mov	ax, 1			; EMM is not present in the system
	pop	ds
	pop	di
	pop	si
	pop	bp
	ret

ASCII_device_name:	db	'EMMXXXX0',0

_check_emm	endp			; end procedure check_emm

	public	_get_page_frame
_get_page_frame	proc	near		; begin procedure get_page_frame
	push	bp

	mov	ah, get_page_segment	; get page frame
	int	EMM_interrupt		; segment address

	or	ah, ah			; if the function failed
	jnz	error_exit1		; return error

	mov	ax, bx			; save page frame segment
	xor	dx, dx
	pop	bp
	ret

error_exit1:
	xor	ax, ax			; return a 0 to signify error
	xor	dx, dx
	pop	bp
	ret

_get_page_frame	endp			; end procedure get_pag_frame

	public	_check_free_pages
_check_free_pages	proc	near	; begin procedure check_free_pages
	push	bp

	mov	ah, free_pages		; get the total number
	int	EMM_interrupt		; of unallocated pages in
					; the system

	or	ah, ah   		; if the function failed
	jnz	error_exit2		; return error

	mov	ax, bx			; save number of free pages
	xor	dx, dx
	pop	bp
	ret

error_exit2:
	xor	ax, ax			; return a 0 to signify error
	xor	dx, dx
	pop	bp
	ret

_check_free_pages	endp		; end procedure check_free_pages

	public	_allocate_pages
_allocate_pages	proc	near		; begin procedure allocate_pages
	push	bp			; save bp
	mov	bp, sp			; set up to get arguments

	mov	ah, get_handle		; get a handle and allocate */
	mov	bx, [bp+04]		; pages to that handle */
	int	EMM_interrupt

	or	ah, ah			; if the function failed
	jnz	error_exit3		; return error

	mov	ax, dx			; save handle
	xor	dx, dx

	pop	bp			; restore bp
	ret

error_exit3:
	xor	ax, ax			; return a 0 to signify error
	xor	dx, dx

	pop	bp			; restore bp
	ret

_allocate_pages	endp			; end procedure allocate_pages

	public	_seek
_seek	proc	near			; begin procedure seek
	push	bp			; save bp
	mov	bp, sp			; set up to get arguments

	xor	cx, cx			; zero cx
	mov	bx, [bp+04]		; get initial window page number
	mov	dx, [bp+10]		; memory segment handle

start_loop:
	mov	al, cl			; put in physical page number
	mov	ah, map_page		; map page function
	int	EMM_interrupt

	or	ah, ah			; non-zero in ah indicates
	jnz	seek_error		; there was an error

	inc	bx                      ; increment logical page
	inc	cx			; increment physical page
	cmp	cl, [bp+08] 		; see if function complete
	jne	start_loop

seek_done:
	xor	ax, ax			; 0 means no error
	pop	bp			; restore bp
	ret

seek_error:
	mov	ax, 1			; 1 means error occured
	pop	bp			; restore bp
	ret

_seek	endp				; end procedure seek

	public	_release
_release	proc	near			; begin procedure release
	push	bp			; save bp
	mov	bp, sp			; set up to get arguments

	mov	dx, [bp+04]		; memory segment handle
	mov	ah, deallocate_pages	; deallocate pages function
	int	EMM_interrupt		; EMM interrupt

	or	ah, ah			; non-zero in ah indicates
	jnz	error_exit4		; there was an error

	xor	ax, ax			; 0 means no error
	pop	bp			; restore bp
	ret

error_exit4:
	mov	ax, 1			; 1 means error occured
	pop	bp			; restore bp
	ret

_release	endp			; end precedure release

	end				; end of assembly, where to start pgm
