/*      main.c   Revision 2.0  12/13/95                             */

/* This software was written by Symbios Logic Inc. to develop and test new
 * products.  Symbios Logic assumes no liability for its use.  This software
 * is released to the public domain to illustrate certain programming
 * techniques for the 53c720 in target mode.
 *
 */

/*
 * Revision 2.0 12/13/95  Scott Rawlings
 * Changed code from polling ISTAT for interrupts to IRQ driven.  Added menu
 * options for greater user control and ease of use.  Streamlined code for
 * faster operation.
 *
 * Revision 1.0 8/23/95  Scott Rawlings
 * Initial release.
 */

#include <dos.h>
#include <bios.h>
#include <mem.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <alloc.h>
#include <math.h>

#include "720targ.out" /* contains the code for the SCSI SCRIPTS */

#include "types.h"
#include "siop.h"
#include "tools.h"
#include "table.h"
#include "tools_lm.c"
#include "emm.h"

#define page_size 16384
#define PIC2_MASK_REG 0xA1 /* slave interrupt cont. mask reg */
#define PIC2_EOI_REG 0xA0 /* end of interrupt register, slave */
#define PIC1_EOI_REG 0x20 /* end of interrupt register, master */
#define IRQ11 0x73 /* vector of IRQ11 */
#define IRQ12 0x74 /* vector of IRQ12 */
#define IRQ11_mask 0xF7	/* mask for enabling IRQ11 */
#define IRQ12_mask 0xEF	/* mask for enabling IRQ12 */

static char     version[] =
"$Id: main.c 2.0 12/13/95 Scott Rawlings $";

/* Prototypes */
static void     init_dma(void);
static void     init_table(void);
static void     init_siop(void);
static void     init_script(void);
static void		 __interrupt isr();
static void		 init_irq(void);
static void		 release_irq(void);

volatile ubyte far *siop_reg; /* pointer to c720's registers */
uword SIOP_SEGMENT_ADDRESS; /* segment in memory where 720 exists */
uquad *script; /* pointer to SCSI SCRIPTs */
table *buffer_table, *dummy_table; /* the tables used for table-indirect */
ubyte medium_buf[49152]; /* drive cache */
uquad EMM_handle; /* handle to reference memory assigned to us */
ubyte old_irq_mask; /* stores the old mask so the interrupt can be released */

ubyte far *medium_window;
#define medium_window_size 65536

/* 8-bit registers */
ubyte dstat, sist0, sist1, istat, scntl3, ssid, sxfer, dcmd, ctest2;

/* current SCSI phase */
ubyte phase;

/* 32-bit registers */
uquad dsp, vector;
ubyte addr[4], dbc[3], dsps[4];

/* other global variables */
int clk, tp, id, wide, irq, busy = 0;
int i, i_max, p, x, first = 1, period, offset, max_offset = 8, error;
int return_code, return_flag = 4, done_synch = 0, done_wide = 0, got_int;
unsigned int data_buf_size, blocks_per_page, block_size;
unsigned int k;
unsigned long int xferp, bytes_left, byte_count, used_pages;
unsigned long int j, l, last_base_l = 0xffffffff;
unsigned long int lba, last_lba = 0xffffffff, last_byte_count = 0;
float cd;

/* SCRIPTS entry points */
uquad wait_select, msg_out_phase, command_phase, tur, request_sense;
uquad read_return, read_reconnect, write_return, write_reconnect;
uquad msg_in_phase, inquiry, read_capacity, copy_data, synch_wide_neg_return;
uquad stopped_busy_wait_select, stopped_busy_tur;

/* SCRIPTS base address pointer */
uquad scripts_address;

/* other buffers */
ubyte command_buf[12];
ubyte dummy_command_buf[12];
ubyte message_out_buf[1];
ubyte dummy_message_out_buf[1];
ubyte sync_neg[2];
ubyte wide_neg[1];

/* pre-set buffers */
static ubyte message_reject[] =
 {
  0x07
 };

static ubyte identify_msg[] =
 {
  0x80
 };

static ubyte command_complete[] =
 {
  0x00
 };

static ubyte device_busy[] =
 {
  0x08
 };

static ubyte status_good[] =
 {
  0x00
 };

static ubyte check_condition[] =
 {
  0x02
 };

static ubyte save_pointers_msg[] =
 {
  0x02
 };

static ubyte restore_pointers_msg[] =
 {
  0x03
 };

static ubyte synch_msg_in_buf[] =
 {
  0x01, 0x03, 0x01, 0x19, 0x08
 };

static ubyte wide_msg_in_buf[] =
 {
  0x01, 0x02, 0x03, 0x00
 };

static ubyte sel_info[] =
 {
  0x00, 0x00, 0x07, 0x33
 };

static ubyte disc_msg[] =
 {
  0x02, 0x04
 };

static ubyte read_capacity_data[] =
 {
  0x00, 0x00, 0x0f, 0xff, 0x00, 0x00, 0x02, 0x00
 };

static ubyte inquiry_data[] =
 {
  0x00, 0x00, 0x02, 0x02, 0x1f, 0x00, 0x00, 0x10, 0x53, 0x59, 0x4d, 0x42,
  0x49, 0x4f, 0x53, 0x20, 0x52, 0x41, 0x4d, 0x20, 0x44, 0x49, 0x53, 0x4b,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x31, 0x2e, 0x30, 0x30
 };

static ubyte bad_lun_inquiry_data[] =
 {
  0x7f, 0x00, 0x02, 0x02, 0x1f, 0x00, 0x00, 0x10, 0x53, 0x59, 0x4d, 0x42,
  0x49, 0x4f, 0x53, 0x20, 0x52, 0x41, 0x4d, 0x20, 0x44, 0x49, 0x53, 0x4b,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x31, 0x2e, 0x30, 0x30
 };

static ubyte no_sense_data[] =
 {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte reset_sense_data[] =
 {
  0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x29, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte unit_not_ready[] =
 {
  0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte illegal_command[] =
 {
  0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x24, 0x00, 0x00, 0x00, 0x00, 0x00
 };

static ubyte seek_error[] =
 {
  0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00, 0x00,
  0x15, 0x00, 0x00, 0x00, 0x00, 0x00
 };

main()
{

 /* local only variables */
 unsigned int new_block_size, lba_offset, value, inc_value, reg;
 unsigned long int use_lba, end_lba, file_size, size;
 uquad page_frame, free_pages;
 char file_name[32], cont;
 FILE *save_file;

 /* initial startup values need to be gotten from user */
 error = check_emm(); /* check to see if emm exists in system */
 if (error)
  {
	printf("EMM not running in system\n");
	exit(1);
  }
 free_pages = check_free_pages(); /* check number of free pages */
 used_pages = free_pages; /* use all of the free pages */
 if (free_pages == 0) /* if there are none, something is worng */
  {
	printf("An EMM error occured\n");
	exit(1);
  }
 size = free_pages * page_size; /* calculate the size */
 printf("Ram Disk Size = %lu (%.2f in MBytes)\n", size, (float)free_pages / 64);
 do
  {
	printf("Bytes Per Logical Block (512, 1024, 2048, 4096) = ?");
	scanf("%lu", &block_size);
  }while ((block_size != 512) && (block_size != 1024)    /* block size is */
		 && (block_size != 2048) && (block_size != 4096)); /* limited to these */
																			/* values */

 end_lba = (size / block_size) - 1; /* last available lba */
 printf("Size of Disk cache (in Kilobytes)(Max = 48k) = ?");
 scanf("%u", &data_buf_size); /* get disk cache size */
 if (data_buf_size > 48)
  {
	printf("size too large, using 48k buffer\n");
	data_buf_size = 48;
  }
 data_buf_size *= 1024;

 /* set up the data for read capacity commands */
 for (i = 0, x = 3; i < 4; i++, x--)
  read_capacity_data[i] = (end_lba >> 8 * x) & 0xff;
 for (i = 0, x = 3; i < 4; i++, x--)
  read_capacity_data[i + 4] = (block_size >> 8 * x) & 0xff;

 page_frame = get_page_frame(); /* get the emm window address */

 /* turn the page frame into a pointer that we can use */
 medium_window = (ubyte far *) MK_FP(page_frame, 0);

 if (medium_window == 0) /* if this happens, something is very wrong */
  {
	printf("An EMM error occured\n");
	exit(1);
  }
 EMM_handle = allocate_pages(free_pages); /* request all free pages */
 if (EMM_handle == 0)
  {
	printf("An EMM error occured\n");
	exit(1);
  }

 /* get the segment address that the 720 card is mapped to */
 do
  {
	printf("Chip Segment Address (d000, d800) = ?");
	scanf("%x", &SIOP_SEGMENT_ADDRESS);
  } while ((SIOP_SEGMENT_ADDRESS != 0xd000)
		  && (SIOP_SEGMENT_ADDRESS != 0xd800));

 printf("SCLK speed in MHz = ?"); /* get SCLK speed */
 scanf("%d", &clk);
 tp = (int)((1 / (float)clk) * 1000); /* calculate the period */
 printf("SCSI ID = ?"); /* get the SCSI id to use */
 scanf("%d", &id);

 /* get the width of the SCSI bus that we are connected to */
 do
  {
	printf("SCSI bus width (0 = narrow, 1 = wide) = ?");
	scanf("%d", &wide);
  } while((wide != 0) && (wide != 1));

 /* if we are on a wide bus, change the inquiry data */
 if (wide)
  {
	inquiry_data[7] = 0x30;
	bad_lun_inquiry_data[7] = 0x30;
  }

 /* find out which interrupt we need to hook */
 do
  {
	printf("IRQ line (11 or 12) = ?");
	scanf("%d", &irq);
  } while((irq != 11) && (irq != 12));

 init_dma();
 init_siop();
 init_table();
 init_script();
 init_irq();

 blocks_per_page = page_size / block_size;

 /* Set up initial values in the table */
 buffer_table[msg_out_buf].count = 1;
 buffer_table[msg_out_buf].address = getPhysAddr(message_out_buf);
 buffer_table[cmd_buf].count = 0;
 buffer_table[cmd_buf].address = getPhysAddr(command_buf);
 buffer_table[synch_neg_msg_out].count = 2;
 buffer_table[synch_neg_msg_out].address = getPhysAddr(sync_neg);
 buffer_table[wide_neg_msg_out].count = 1;
 buffer_table[wide_neg_msg_out].address = getPhysAddr(wide_neg);
 buffer_table[stat_buf].count = 1;
 buffer_table[stat_buf].address = getPhysAddr(check_condition);
 buffer_table[identify_msg_in_buf].count = 1;
 buffer_table[identify_msg_in_buf].address = getPhysAddr(identify_msg);
 buffer_table[msg_in_buf].count = 1;
 buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
 buffer_table[save_pointers].count = 1;
 buffer_table[save_pointers].address = getPhysAddr(save_pointers_msg);
 buffer_table[disconnect_msg].count = 2;
 buffer_table[disconnect_msg].address = getPhysAddr(disc_msg);
 buffer_table[sense_data_buf].address = getPhysAddr(reset_sense_data);
 buffer_table[capacity_data_buf].count = 8;
 buffer_table[capacity_data_buf].address = getPhysAddr(read_capacity_data);
 buffer_table[data_buf].address = getPhysAddr(medium_buf);

 dummy_table[msg_out_buf].count = 1;
 dummy_table[msg_out_buf].address = getPhysAddr(dummy_message_out_buf);
 dummy_table[cmd_buf].count = 0;
 dummy_table[cmd_buf].address = getPhysAddr(dummy_command_buf);
 dummy_table[stat_buf].count = 1;
 dummy_table[stat_buf].address = getPhysAddr(device_busy);
 dummy_table[msg_in_buf].count = 1;
 dummy_table[msg_in_buf].address = getPhysAddr(command_complete);
 dummy_table[sense_data_buf].address = getPhysAddr(no_sense_data);
 dummy_table[capacity_data_buf].count = 8;
 dummy_table[capacity_data_buf].address = getPhysAddr(read_capacity_data);

 /* set base scripts address */
 scripts_address = getPhysAddr(script);

 /* assign the entry point addresses */
 wait_select = scripts_address + Ent_wait_select;
 msg_out_phase = scripts_address + Ent_msg_out_phase;
 tur = scripts_address + Ent_tur;
 stopped_busy_tur = scripts_address + Ent_stopped_busy_tur;
 request_sense = scripts_address + Ent_request_sense;
 read_return = scripts_address + Ent_read_return;
 read_reconnect = scripts_address + Ent_read_reconnect;
 write_return = scripts_address + Ent_write_return;
 write_reconnect = scripts_address + Ent_write_reconnect;
 synch_wide_neg_return = scripts_address + Ent_synch_wide_neg_return;
 msg_in_phase = scripts_address + Ent_msg_in_phase;
 stopped_busy_wait_select = scripts_address + Ent_stopped_busy_wait_select;
 inquiry = scripts_address + Ent_inquiry;
 read_capacity = scripts_address + Ent_read_capacity;

 /* Start first script */
 write_long_reg(DSP, wait_select);

 /* clear the screen and print the user menu */
 clrscr();
 printf("Main Menu:\n");
 printf("1. Save Ram Disk To Hard Drive\n");
 printf("2. Restore Ram Disk From Hard Drive\n");
 printf("3. Print A Logical Block To The Screen\n");
 printf("4. Modify A Logical Block (Not Recommended In Normal Operation)\n");
 printf("5. Read A Chip Register\n");
 printf("6. Modify A Chip Register (Not Recommended In Normal Operation)\n");
 printf("7. Quit\n");
 printf("Selecting Any Of These Options Will Halt SCRIPTS\n");

 /* as long as no menu key is pressed, this loop will simply keep enabling */
 /* system interupts.  if a key is pressed, we go into the case statement to */
 /* determine which option was chosen and handle it.  the values that had to */
 /* be used in the case statement are keyboard scan codes */
 do
  {
	enable();
	if (_bios_keybrd(_NKEYBRD_READY) != 0)
	 {
	  got_int = 0;
	  siop_reg[ISTAT] |= 0x20; /* set the SIGP bit in ISTAT */
	  printf("\nWriting SIGP Bit To Abort SCRIPTS....\n");

	  /* loop until the interrupt happens */
	  do
		{
		 enable();
		} while (!got_int);
	  ctest2 = siop_reg[CTEST2]; /* clear the SIGP bit */
	  printf("SCRIPTS Successfully Aborted\n\n");

	  /* determine which key was pressed */
	  switch(_bios_keybrd(_NKEYBRD_READ))
		{

		 /* save ram disk from memory to a hard drive */
		 case 561: printf("Enter File Name With Path:");
					  scanf("%s", &file_name); /* get file name */

					  /* open file.  on error, print error message and break */
					  if ((save_file = fopen(file_name, "wb")) == NULL)
						{
						 printf("Error Opening File, Save Not Possible\n");
						 printf("Press Any Key To Continue\n");
						 cont = getch();
						 break;
						}
					  printf("This Will Take A Few Minutes\n");

					  /* save the ram disk size and block size */
					  fprintf(save_file, "%lu %u", size, block_size);

					  /* we can move four emm pages at a time, for a total of 64K */
					  /* per window move.  the for loop starts at page zero and */
					  /* increments by four until the last group of four is moved */
					  for(l = 0; l < used_pages; l + 4)
						{

						 /* this ensures that we don't try to map physical windows */
						 /* to virtual memory space beyond our allocation */
						 p = used_pages - l;

						 /* if there are more than four pages til the end of our */
						 /* allocation, then we use four */
						 if (p > 4)
						  p = 4;

						 /* seek to the next set of logical pages */
						 error = seek(l, p, EMM_handle);

						 /* error during seek is bad */
						 if (error)
						  {
							printf("error %x in seek to logical page %x\n", error, l);
							error = release(EMM_handle);
							if (error)
							 printf("error deallocating pages\n");
							exit(1);
						  }

						 /* move data from the medium window to the hard drive */
						 for (j = 0; j < (p * page_size); j++)
						  fprintf(save_file, "%c", medium_window[j]);
						}
					  fclose(save_file); /* close file */
					  printf("Ram Disk Saved\n");
					  printf("Press Any Key To Continue\n");
					  cont = getch();
					  break;

		 /* restore ram disk from a hard drive to memory */
		 case 818: printf("Enter File Name With Path:");
					  scanf("%s", &file_name); /* get file name */

					  /* open file.  on error, print error message and break */
					  if ((save_file = fopen(file_name, "rb")) == NULL)
						{
						 printf("Error Opening File, Restore Not Possible\n");
						 printf("Press Any Key To Continue\n");
						 cont = getch();
						 break;
						}

					  /* get ram disk size and block size */
					  fscanf(save_file, "%lu %u", &file_size, &new_block_size);
					  if (file_size > size) /* if file too large, print error */
						{                    /* and break */
						 printf("File Too Large, Aborting Load\n");
						 fclose(save_file);
						 printf("Press Any Key To Continue\n");
						 cont = getch();
						 break;
						}

					  /* if the file size or block size is different than the */
					  /* current setup, the setup must be modified */
					  if ((file_size != size) || (new_block_size != block_size))
						{
						 size = file_size;
						 block_size = new_block_size;
						 end_lba = (size / block_size) - 1;
						 used_pages = (end_lba / blocks_per_page) + 1;
						 printf("New Drive Size = %.2f MBytes (%ld Bytes)\n",
															 (((float)size) / 1048576), size);
						 printf("New Block Size = %u\n", block_size);
						 for (i = 0, x = 3; i < 4; i++, x--)
						  read_capacity_data[i] = (end_lba >> 8 * x) & 0xff;
						 for (i = 0, x = 3; i < 4; i++, x--)
						  read_capacity_data[i + 4] = (block_size >> 8 * x) & 0xff;
						}
					  printf("This Will Take A Few Minutes\n");

					  /* we can move four emm pages at a time, for a total of 64K */
					  /* per window move.  the for loop starts at page zero and */
					  /* increments by four until the last group of four is moved */
					  for(l = 0; l < used_pages ; l + 4)
						{

						 /* this ensures that we don't try to map physical windows */
						 /* to virtual memory space beyond our allocation */
						 p = used_pages - l;

						 /* if there are more than four pages til the end of our */
						 /* allocation, then we use four */
						 if (p > 4)
						  p = 4;

						 /* seek to the next set of logical pages */
						 error = seek(l, p, EMM_handle);

						 /* error during seek is bad */
						 if (error)
						  {
							printf("error %x in seek to logical page %x\n", error, l);
							error = release(EMM_handle);
							if (error)
							 printf("error deallocating pages\n");
							exit(1);
						  }

						 /* move data from the hard drive to the medium window */
						 for (j = 0; j < (p * page_size); j++)
						  fscanf(save_file, "%c", &medium_window[j]);
						}
					  fclose(save_file); /* close file */
					  printf("Ram Disk Restored\n");
					  printf("Host Should Be Reset For Proper Operation\n");
					  printf("Press Any Key To Continue\n");
					  cont = getch();
					  break;

		 /* print a logical block to the screen */
		 case 1075:printf("Highest Logical Block = %lu\n", end_lba);
					  do
						{
						 printf("Enter Logical Block Number:");
						 scanf("%lu", &use_lba); /* get block number */
						}while (use_lba > end_lba); /* must be in range */

					  l = use_lba / blocks_per_page; /* calculate the logical page */
					  last_base_l = l; /* stores what block was last put in cache */

					  /* this ensures that we don't try to map physical windows */
					  /* to virtual memory space beyond our allocation */
					  p = used_pages - l;

					  /* if there are more than four pages til the end of our */
					  /* allocation, then we use four */
					  if (p > 4)
						p = 4;

					  /* seek to the next set of logical pages */
					  error = seek(l, p, EMM_handle);

					  /* error during seek is bad */
					  if (error)
						{
						 printf("error %xh in seek to logical page %xh\n", error, l);
						 error = release(EMM_handle);
						 if (error)
						  printf("error deallocating pages\n");
						 exit(1);
						}

					  /* clear the screen and print the logical block */
					  clrscr();

					  /* calculate the relative starting point within the window */
					  i = ((use_lba * block_size)
											 - ((use_lba / blocks_per_page) * page_size));
					  i_max = i + block_size; /* the end of the lba */
					  do
						{
						 for (p = 0; p < 16; p++) /* 16 lines per page */
						  {
							for (x = 0; x < 16; x++) /* 16 bytes per line */
							 printf("%02x ", medium_window[i + x]); /* in hex */
							printf("     ");

							/* if this isn't a printable character, print a "." */
							for (x = 0; x < 16; x++, i++)
							 if ((medium_window[i] < 32)
							  || (medium_window[i] > 126))
							  printf(".");
							 else
							  printf("%c", medium_window[i]); /* in ascii */
							printf("\n");
						  }

						 /* stop after each full page */
						 printf("Press Any Key To Continue\n");
						 cont = getch();
						} while (i < i_max);
					  break;

		 /* modify a logical block */
		 case 1332:printf("Highest Logical Block = %lu\n", end_lba);
					  do
						{
						 printf("Enter Logical Block Number:");
						 scanf("%lu", &use_lba); /* get block number */
						}while (use_lba > end_lba); /* must be in range */
					  do
						{
						 printf("Enter Start Offset:");
						 scanf("%u", &lba_offset); /* get start offset in block */
						}while (lba_offset > block_size - 1); /* must be in block */
					  do
						{
						 printf("Enter Byte Count To Modify:");
						 scanf("%u", &byte_count); /* get number of bytes to mod */

						/* can't cross logical block boundary */
						}while (byte_count > (block_size - lba_offset));

					  /* if only one byte, get new value */
					  if (byte_count == 1)
						{
						 do
						  {
							printf("Enter New Value (In Hex):");
							scanf("%x", &value);
						  }while (value > 0xff);
						}

					  /* if more than one byte, get start and increment values */
					  if (byte_count > 1)
						{
						 do
						  {
							printf("Enter Starting Value (In Hex):");
							scanf("%x", &value);
						  }while (value > 0xff);
						 printf("Enter Increment Value (In Hex):");
						 scanf("%x", &inc_value);
						}

					  l = use_lba / blocks_per_page; /* calculate the logical page */
					  last_base_l = l; /* stores what block was last put in cache */

					  /* this ensures that we don't try to map physical windows */
					  /* to virtual memory space beyond our allocation */
					  p = used_pages - l;

					  /* if there are more than four pages til the end of our */
					  /* allocation, then we use four */
					  if (p > 4)
						p = 4;

					  /* seek to the next set of logical pages */
					  error = seek(l, p, EMM_handle);

					  /* error during seek is bad */
					  if (error)
						{
						 printf("error %x in seek to logical page %x\n", error, l);
						 error = release(EMM_handle);
						 if (error)
						  printf("error deallocating pages\n");
						 exit(1);
						}

					  /* calculate the relative starting point within the window */
					  i = ((use_lba * block_size)
						 - ((use_lba / blocks_per_page) * page_size))
						 + lba_offset;
					  i_max = i + byte_count; /* the end of the range to modify */

					  /* set up data in buffer */
					  for ( ; i < i_max; i++)
						{
						 medium_window[i] = (ubyte)value;
						 value += inc_value;
						}
					  printf("Logical Block Modified\n");
					  printf("Press Any Key To Continue\n");
					  cont = getch();
					  break;

		 /* read a chip register */
		 case 1589:printf("Enter Register Number (In Hex):");
					  scanf("%x", &reg); /* get register number */
					  printf("\nRegister %02xh = %02xh\n", reg, siop_reg[reg]);
					  printf("Press Any Key To Continue\n");
					  cont = getch();
					  break;

		 /* modify a chip register */
		 case 1846:printf("Enter Register Number (In Hex):");
					  scanf("%x", &reg); /* get register number */

					  /* print out the old register value */
					  printf("Old Value = %02xh\n", siop_reg[reg]);
					  do
						{
						 printf("Enter New Value (In Hex):");
						 scanf("%x", &value); /* get the new value */
						}while (value > 0xff);
					  siop_reg[reg] = value;
					  printf("\nRegister Modified\n");
					  printf("Press Any Key To Continue\n");
					  cont = getch();
					  break;

		 /* quit */
		 case 2103:release_irq(); /* release interrupt and restore mask */
					  error = release(EMM_handle);
					  if (error)
						printf("error deallocating pages\n");
					  exit(0); /* terminate program */
					  break;
		}

	  /* restart SCRIPTS, clear the screen, and reprint the user menu */
	  write_long_reg(DSP, wait_select);
	  clrscr();
	  printf("Main Menu:\n");
	  printf("1. Save Ram Disk To Hard Drive\n");
	  printf("2. Restore Ram Disk From Hard Drive\n");
	  printf("3. Print A Logical Block To The Screen\n");
	  printf("4. Modify A Logical Block (Not Recommended In Normal ");
	  printf("Operation)\n");
	  printf("5. Read A Chip Register\n");
	  printf("6. Modify A Chip Register (Not Recommended In Normal ");
	  printf("Operation)\n");
	  printf("7. Quit\n");
	  printf("Selecting Any Of These Options Will Halt SCRIPTS\n");
	 }
  } while (1); /* loop until the quit option is invoked */
}

/* interrupt service routine invoked by the chip interrupt pin */
static void
__interrupt isr()
{
 disable(); /* disable system interrupts */

 /* Determine what caused the interrupt and deal with it appropriately */
 istat = siop_reg[ISTAT];
 if (istat & 1) /* this is a DMA interrupt */
  {
	dstat = siop_reg[DSTAT] & 0x7f;
	switch (dstat)
	 {
	  case 0x04: /* scripts interrupt instruction */
		vector = (((uquad)siop_reg[DSPS + 3] << 24)
				  | ((uquad)siop_reg[DSPS + 2] << 16)
				  | ((uquad)siop_reg[DSPS + 1] << 8)
				  | (siop_reg[DSPS]));
		switch (vector)
		 {
		  case 0x00: /* access medium for read */
			if (first) /* this is the first interrupt of the read command */
			 {
			  if (command_buf[0] == 0x08) /* this is a 6-byte read command */
				{

				 /* translate blocks to bytes */
				 bytes_left = ((unsigned long int)command_buf[4]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = ((((unsigned long int)command_buf[1] & 0x1f) << 16)
						 | ((unsigned long int)command_buf[2] << 8)
						 | command_buf[3]);
				}
			  else /* this is a 10-byte read command */
				{

				 /* translate blocks to bytes */
				 bytes_left = (((unsigned long int)command_buf[7] << 8)
									| command_buf[8]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = (((unsigned long int)command_buf[2] << 24)
						| ((unsigned long int)command_buf[3] << 16)
						| ((unsigned long int)command_buf[4] << 8)
						| command_buf[5]);
				}
			  first = 0; /* clear the first flag */
			 }

			/* determine if a portion of the total job, or the entire rest of */
			/* the job will be transfered this time */
			if (bytes_left <= data_buf_size)
			 byte_count = bytes_left;
			else
			 byte_count = data_buf_size;

			/* put byte count in table */
			buffer_table[data_buf].count = byte_count;

			/* check to see if this data is already in the cache */
			if ((lba == last_lba) && (byte_count == last_byte_count))

			 /* if it is, there is no need to move it there again */
			 bytes_left -= byte_count;

			/* the data isn't in the cache, so we have to go get it */
			else
			 {
			  l = (lba / blocks_per_page); /* calculate the logical page */

			  /* if the last logical page accessed is the same as the wanted */
			  /* one, then we don't need to seek */
			  if (l != last_base_l)
				{
				 last_base_l = l; /* change the last page accessed */

				 /* this ensures that we don't try to map physical windows */
				 /* to virtual memory space beyond our allocation */
				 p = used_pages - l;

				 /* if there are more than four pages til the end of our */
				 /* allocation, then we use four */
				 if (p > 4)
				  p = 4;

				 /* seek to the wanted logical page */
				 error = seek(l, p, EMM_handle);
				 if (error) /* error during seek is bad */
				  {
					printf("error %x in seek to lba %x\n", error, lba);
					error = release(EMM_handle);
					if (error)
					 printf("error deallocating pages\n");
					exit(1);
				  }
				}

			  /* calculate the offset into the logical page where the lba begins */
			  j = ((lba * block_size) - ((lba / blocks_per_page) * page_size));

			  /* copy the data from the logical page to the cache */
			  _fmemcpy(medium_buf, (medium_window + j), byte_count);
			  bytes_left -= byte_count; /* update the number of bytes left */
			 }
			if (bytes_left == 0) /* we are done with this command */
			 {
			  last_lba = lba; /* store which lba is at the start of the cache */

			  /* store how many valid bytes exist in the cache */
			  last_byte_count = byte_count;
			  first = 1; /* restore the first flag */

			  /* clear 'not done' in the finished flag */
			  siop_reg[SCRATCHB1] = 0x00;
			 }
			else /* we have more data to transfer */
			 {
			  lba += data_buf_size / block_size; /* increment the lba pointer */

			  /* set 'not done' in the finished flag */
			  siop_reg[SCRATCHB1] = 0x01;
			 }

			/* if semaphore bit set, we are disconnected */
			if (istat & 0x10)
			 {
			  siop_reg[ISTAT] = 0x00; /* clear the semaphore bit */
			  write_long_reg(DSP, read_reconnect); /* start SCRIPT */
			 }
			else
			 write_long_reg(DSP, read_return); /* start SCRIPT */
			break;

		  case 0x01: /* access medium for write */
			if (first) /* this is the first interrupt of the write command */
			 {
			  if (command_buf[0] == 0x0a) /* this is a 6-byte write command */
				{

				 /* translate blocks to bytes */
				 bytes_left = ((unsigned long int)command_buf[4]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = ((((unsigned long int)command_buf[1] & 0x1f) << 16)
						 | ((unsigned long int)command_buf[2] << 8)
						 | command_buf[3]);
				}
			  else  /* this is a 10-byte write command */
				{

				 /* translate blocks to bytes */
				 bytes_left = (((unsigned long int)command_buf[7] << 8)
									| command_buf[8]) * block_size;

				 /* calculate the logical block address to be accessed */
				 lba = (((unsigned long int)command_buf[2] << 24)
						| ((unsigned long int)command_buf[3] << 16)
						| ((unsigned long int)command_buf[4] << 8)
						| command_buf[5]);
				}

			  /* determine if a portion of the total job, or the entire rest of */
			  /* the job will be transfered this time */
			  if (bytes_left <= data_buf_size)
				byte_count = bytes_left;
			  else
				byte_count = data_buf_size;

			  /* put byte count in table */
			  buffer_table[data_buf].count = byte_count;
			  first = 0; /* clear the first flag */
			 }
			else
			 {
			  l = (lba / blocks_per_page); /* calculate the logical page */

			  /* if the last logical page accessed is the same as the wanted */
			  /* one, then we don't need to seek */
			  if (l != last_base_l)
				{
				 last_base_l = l; /* change the last page accessed */

				 /* this ensures that we don't try to map physical windows */
				 /* to virtual memory space beyond our allocation */
				 p = used_pages - l;

				 /* if there are more than four pages til the end of our */
				 /* allocation, then we use four */
				 if (p > 4)
				  p = 4;

				 /* seek to the wanted logical page */
				 error = seek(l, p, EMM_handle);
				 if (error) /* error during seek is bad */
				  {
					printf("error %x in seek to lba %x\n", error, lba);
					error = release(EMM_handle);
					if (error)
					 printf("error deallocating pages\n");
					exit(1);
				  }
				}

			  /* calculate the offset into the logical page where the lba begins */
			  j = ((lba * block_size) - ((lba / blocks_per_page) * page_size));

			  /* copy the data from the cache to the logical page */
			  _fmemcpy((medium_window + j), medium_buf, byte_count);
			  bytes_left -= byte_count; /* update the number of bytes left */
			  if (bytes_left == 0) /* we are done with this command */
				{
				 last_lba = lba; /* store which lba is at the start of the cache */

				 /* store how many valid bytes exist in the cache */
				 last_byte_count = byte_count;
				 first = 1; /* restore the first flag */

				 /* clear 'not done' in the finished flag */
				 siop_reg[SCRATCHB1] = 0x00;
				}
			  else /* we have more data to transfer */
				{

				 /* determine if a portion of the total job, or the entire rest */
				 /* of the job will be transfered this time */
				 if (bytes_left <= data_buf_size)
				  byte_count = bytes_left;
				 else
				  byte_count = data_buf_size;

				 /* put byte count in table */
				 buffer_table[data_buf].count = byte_count;
				 lba += data_buf_size / block_size; /* increment the lba pointer */

				 /* set 'not done' in the finished flag */
				 siop_reg[SCRATCHB1] = 0x01;
				}
			 }

			/* if semaphore bit set, we are disconnected */
			if (istat & 0x10)
			 {
			  siop_reg[ISTAT] = 0x00; /* clear the semaphore bit */
			  write_long_reg(DSP, write_reconnect); /* start SCRIPT */
			 }
			else
			 write_long_reg(DSP, write_return); /* start SCRIPT */
			break;

		  case 0x02: /* this is a seek command */
			if (command_buf[0] == 0x0b) /* 6-byte seek command */
			 lba = ((((unsigned long int)command_buf[1] & 0x1f) << 16)
					 | ((unsigned long int)command_buf[2] << 8)
					 | command_buf[3]);
			else /* ten byte seek command */
			 lba = (((unsigned long int)command_buf[2] << 24)
					| ((unsigned long int)command_buf[3] << 16)
					| ((unsigned long int)command_buf[4] << 8)
					| command_buf[5]);
			l = (lba / blocks_per_page); /* calculate the logical page */

			/* this ensures that we don't try to map physical windows */
			/* to virtual memory space beyond our allocation */
			p = used_pages - l;

			/* if there are more than four pages til the end of our */
			/* allocation, then we use four */
			if (p > 4)
			 p = 4;

			/* seek to the wanted logical page */
			error = seek(l, p, EMM_handle);
			if (!error)

			 /* if sucessful, send back good status */
			 buffer_table[stat_buf].address = getPhysAddr(status_good);
			else
			 {

			  /* if the seek failed, send back check condition and set up */
			  /* seek error in the sense data buffer */
			  buffer_table[stat_buf].address = getPhysAddr(check_condition);
			  buffer_table[sense_data_buf].address = getPhysAddr(seek_error);

			  /* says that after command complete, status should go back to good */
			  return_flag = 2;
			 }
			write_long_reg(DSP, tur); /* start SCRIPT */
			break;

		  case 0x03: /* synchronous negotiation occured */
			cd = 1; /* clock dividor for scntl3 register */
			period = ((int)sync_neg[0]); /* get period sent by the initiator */
			offset = ((int)sync_neg[1]); /* get offset sent by the initiator */
			if (period <= tp) /* if their fastest is faster than ours */
			 {
			  synch_msg_in_buf[3] = (ubyte)tp; /* our fastest period */
			  xferp = 0; /* xferp for sxfer register */
			 }

			/* we are faster, we need to set up to run at the initiator's rate */
			else
			 {
			  synch_msg_in_buf[3] = ((ubyte)period); /* return period received */

			  /* calculate the xferp for the current clock divider.  if the */
			  /* xferp is too high, change the clock divider and try again until */
			  /* either a working combination is found or it is determined that */
			  /* synchronous is not possible */
			  xferp = ((period * 4) / (tp * cd) - 4);
			  if (xferp > 7)
				{
				 cd = 1.5;
				 xferp = ((period * 4) / (tp * cd) - 4);
				 if (xferp > 7)
				  {
					cd = 2;
					xferp = ((period * 4) / (tp * cd) - 4);
					if (xferp > 7)
					 {
					  cd = 3;
					  xferp = ((period * 4) / (tp * cd) - 4);

					  /* if xferp is greater than 7 on our last divider, we */
					  /* cannot do synchronous with this initiator */
					  if (xferp > 7)
						{
						 buffer_table[msg_in_buf].address =
																	getPhysAddr(message_reject);
						 write_long_reg(DSP, msg_in_phase);
						 return_code = 1; /* rejecting a bad message */
						 xferp = 10; /* 10 means we rejected synch neg */
						}
					 }
				  }
				}
			 }
			if (xferp == 10) /* we rejected the synch neg message */
			 offset = 0; /* go asynchronously */
			else
			 {
			  if (offset > max_offset) /* initiator request is too big */
				offset = max_offset; /* send back our best offset */
			  synch_msg_in_buf[4] = ((ubyte)offset); /* set offset in buffer */

			  /* switch cannot use float, so convert to int */
			  if (cd == 1.5)
				cd = 4;
			  x = ((int)cd);

			  /* zero out the synchronous clock divider in SCTNL3 */
			  siop_reg[SCNTL3] &= 0x0f;

			  switch (x) /* set up the clock divider */
				{
				 case 1: siop_reg[SCNTL3] |= 0x10;
							break;
				 case 2: siop_reg[SCNTL3] |= 0x30;
							break;
				 case 3: siop_reg[SCNTL3] |= 0x40;
							break;
				 case 4: siop_reg[SCNTL3] |= 0x20;
							break;
				}
			  scntl3 = siop_reg[SCNTL3]; /* get the reg value to calculate speed */
			  sxfer = ((((ubyte)xferp) << 5) | offset); /* set up sxfer */

			  /* after synchronous negotiation, we must update the table info */
			  sel_info[3] = scntl3;
			  sel_info[1] = sxfer;
			  buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

			  siop_reg[SXFER] = sxfer; /* write the sxfer register */
			  buffer_table[neg_msg_in].count = 5; /* number of message bytes */

			  /* put address of the negotiation message in the table */
			  buffer_table[neg_msg_in].address = getPhysAddr(synch_msg_in_buf);

			  /* if we haven't done wide negotiation yet, clear the screen and */
			  /* reprint the user menu before giving the synch results */
			  if (!done_wide)
				{
				 clrscr();
				 printf("Main Menu:\n");
				 printf("1. Save Ram Disk To Hard Drive\n");
				 printf("2. Restore Ram Disk From Hard Drive\n");
				 printf("3. Print A Logical Block To The Screen\n");
				 printf("4. Modify A Logical Block (Not Recommended In ");
				 printf("Normal Operation)\n");
				 printf("5. Read A Chip Register\n");
				 printf("6. Modify A Chip Register (Not Recommended In ");
				 printf("Normal Operation)\n");
				 printf("7. Quit\n");
				 printf("Selecting Any Of These Options Will Halt SCRIPTS\n");
				}
			  printf("\nSynchronous Negotiation Occured\n");

			  /* print out the theoretical maximum transfer rate */
			  printf("Max Transfer Rate = %.2f MBytes/Sec\n",
												 ((1 / ((((1 / (float)clk) * 1000) * cd)
												  * ((float)xferp + 4))) * 1000)
												  * ((float)((scntl3 & 0x08) >> 3) + 1));
			  write_long_reg(DSP, synch_wide_neg_return); /* start SCRIPT */
			  done_synch = 1; /* set the done synch flag */
			 }
			break;

		  case 0x04: /* wide negotiation occured */

			/* if we haven't done synch negotiation yet, clear the screen and */
			/* reprint the user menu before giving the wide negotiation results */
			if (!done_synch)
			 {
			  clrscr();
			  printf("Main Menu:\n");
			  printf("1. Save Ram Disk To Hard Drive\n");
			  printf("2. Restore Ram Disk From Hard Drive\n");
			  printf("3. Print A Logical Block To The Screen\n");
			  printf("4. Modify A Logical Block (Not Recommended In ");
			  printf("Normal Operation)\n");
			  printf("5. Read A Chip Register\n");
			  printf("6. Modify A Chip Register (Not Recommended In ");
			  printf("Normal Operation)\n");
			  printf("7. Quit\n");
			  printf("Selecting Any Of These Options Will Halt SCRIPTS\n");
			 }

			/* narrow request or narrow chip */
			if ((wide_neg[0] == 0x00) | (wide == 0))
			 {
			  printf("\nWide Negotiation Occured\n");
			  printf("Running Narrow\n");
			  wide_msg_in_buf[3] = 0x00; /* set narrow in buffer */
			  buffer_table[neg_msg_in].count = 4; /* number of message bytes */

			  /* put address of the negotiation message in the table */
			  buffer_table[neg_msg_in].address = getPhysAddr(wide_msg_in_buf);
			  write_long_reg(DSP, synch_wide_neg_return); /* start SCRIPT */
			 }
			else

			 /* wide or double wide request*/
			 if ((wide_neg[0] == 0x01) | (wide_neg[0] == 0x02))
			  {
				printf("\nWide Negotiation Occured\n");
				printf("Running Wide\n");
				wide_msg_in_buf[3] = 0x01; /* set wide in buffer */
				buffer_table[neg_msg_in].count = 4; /* number of message bytes */

				/* put address of the negotiation message in the table */
				buffer_table[neg_msg_in].address = getPhysAddr(wide_msg_in_buf);
				siop_reg[SCNTL3] |= 0x08; /* set wide SCSI bit */
				scntl3 = siop_reg[SCNTL3]; /* get reg value to calculate speed */

				/* after wide negotiation, we must update the table info */
				sel_info[3] = scntl3;
				buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

				/* if synchronous negotiation has already occured, update the */
				/* theoretical maximum transfer rate */
				if (done_synch)
				 printf("\nMax Transfer Rate = %.2f MBytes/Sec\n",
												 ((1 / ((((1 / (float)clk) * 1000) * cd)
												  * ((float)xferp + 4))) * 1000)
												  * ((float)((scntl3 & 0x08) >> 3) + 1));
				write_long_reg(DSP, synch_wide_neg_return); /* start SCRIPT */
				done_wide = 1; /* set the done wide flag */
			  }
			 else /* invalid wide negotiation message */
			  {

				/* set up the message reject in the table */
				buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
				write_long_reg(DSP, msg_in_phase); /* start SCRIPT */
				return_code = 1; /* says this was rejecting a bad message */
			  }
			break;

		  case 0x05: /* a message not handled by scripts was received */
			switch(message_out_buf[0])
			 {
			  case 0x0c: /* bus device reset message */

				/* call the function to reset the chip and set up initial values */
				init_siop();

				/* restore the DSA register */
				write_long_reg(DSA, getPhysAddr(buffer_table));

				/* put the check condition in the status buffer */
				buffer_table[stat_buf].address = getPhysAddr(check_condition);

				/* put command complete into the message in buffer */
				buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
				return_flag = 2; /* we sent a check condition */

				/* put the unit attention sense data in the sense data buffer */
				buffer_table[sense_data_buf].address =
																 getPhysAddr(reset_sense_data);
				write_long_reg(DSP, wait_select); /* start SCRIPT */

				/* return needed variables to their initial values */
				first = 1;
				done_synch = 0;
				done_wide = 0;
				break;

			  case 0x05: /* initiator detected error message */

				/* send back a restore pointers message to retry the transaction */
				buffer_table[msg_in_buf].address =
															getPhysAddr(restore_pointers_msg);

				/* after sending the message, resume the last operation */
				return_code = 2;
				write_long_reg(DSP, msg_in_phase); /* start SCRIPT */
				break;

			  case 0x07: /* we got a message reject -should never happen- */
				release_irq(); /* release interrupt and restore mask */
				error = release(EMM_handle);
				if (error)
				 printf("error deallocating pages\n");
				exit(1); /* terminate program with error code */
				break;

			  case 0x09: /* we got a message parity error message */
				buffer_table[msg_in_buf].address =
															getPhysAddr(restore_pointers_msg);

				/* after sending the message, resume the last operation */
				return_code = 2;
				write_long_reg(DSP , msg_in_phase); /* start SCRIPT */
				break;

			  default: /* any other message gets a message reject */
				buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
				write_long_reg(DSP, msg_in_phase); /* start SCRIPT */

				/* after sending the message, return to message out phase */
				return_code = 1;
				break;
			 }
			break;

		  case 0x06: /* a bad extended message was received */

			/* put message reject in the table */
			buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
			write_long_reg(DSP, msg_in_phase); /* start SCRIPT */

			/* after sending the message, return to message out phase */
			return_code = 1;
			break;

		  case 0x07: /* we sent a message to the initiator */
			switch (return_code)
			 {
			  case 1: /* this was rejecting a bad message */

				/* restore the command complete message in the table */
				buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
				write_long_reg(DSP, msg_out_phase); /* start SCRIPT */
				break;

			  case 2: /* this was from a parity error */

				/* restore the command complete message in the table */
				buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
				write_long_reg(DSP, dsp); /* start SCRIPT */
				break;
			 }
			break;

		  case 0x08: /* got a request sense command */
			if (!busy)
			 {

			  /* if the initiator asked for less sense data than we have, send */
			  /* back only the amount asked for */
			  if (command_buf[4] <= 0x12)
				buffer_table[sense_data_buf].count = command_buf[4];

			  /* if the initiator asked for more sense data than we have, send */
			  /* back only what we have */
			  else
				buffer_table[sense_data_buf].count = 0x12;

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else
			 {

			  /* if the initiator asked for less sense data than we have, send */
			  /* back only the amount asked for */
			  if (dummy_command_buf[4] <= 0x12)
				dummy_table[sense_data_buf].count = dummy_command_buf[4];

			  /* if the initiator asked for more sense data than we have, send */
			  /* back only what we have */
			  else
				dummy_table[sense_data_buf].count = 0x12;

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			 }
			write_long_reg(DSP, request_sense); /* start SCRIPT */
			break;

		  case 0x09: /* got an inquiry command */
			if (!busy)
			 {

			  /* find out what the initiator's id is and put that in the table */
			  /* this assumes that an inquiry will be done before the initiator */
			  /* attempts to transfer data with disconnects and that only one */
			  /* initiator is operating on the SCSI bus */
			  ssid = siop_reg[SSID] & 0x0f; /* get the initiator's id */
			  sel_info[2] = ssid;
			  buffer_table[selector_id].count =
													 (((unsigned long int)sel_info[3] << 24)
													| ((unsigned long int)sel_info[2] << 16)
													| ((unsigned long int)sel_info[1] << 8)
													| sel_info[0]);

			  /* if the initiator asked for less inquiry data than we have, send */
			  /* back only the amount asked for */
			  if (command_buf[4] <= 0x24)
				buffer_table[inquiry_data_buf].count = command_buf[4];

			  /* if the initiator asked for more inquiry data than we have, send */
			  /* back only what we have */
			  else
				buffer_table[inquiry_data_buf].count = 0x24;

			  /* if this command is for a valid lun, send back the normal data */
			  if ((message_out_buf[0] & 0x07) == 0)
				buffer_table[inquiry_data_buf].address = getPhysAddr(inquiry_data);

			  /* if this command is for an invalid lun, send back bad lun data */
			  else
				buffer_table[inquiry_data_buf].address =
														  getPhysAddr(bad_lun_inquiry_data);

			  /* if a check condition exists, we must still send good status */
			  if (return_flag == 2)
				{
				 buffer_table[stat_buf].address = getPhysAddr(status_good);

				 /* after command complete, return status to check condition */
				 return_flag=3;
				}

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else
			 {
			  /* if the initiator asked for less inquiry data than we have, send */
			  /* back only the amount asked for */
			  if (dummy_command_buf[4] <= 0x24)
				dummy_table[inquiry_data_buf].count = dummy_command_buf[4];

			  /* if the initiator asked for more inquiry data than we have, send */
			  /* back only what we have */
			  else
				dummy_table[inquiry_data_buf].count = 0x24;

			  /* if this command is for a valid lun, send back the normal data */
			  if ((message_out_buf[0] & 0x07) == 0)
				dummy_table[inquiry_data_buf].address = getPhysAddr(inquiry_data);

			  /* if this command is for an invalid lun, send back the bad lun data */
			  else
				dummy_table[inquiry_data_buf].address =
														  getPhysAddr(bad_lun_inquiry_data);

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			 }
			write_long_reg(DSP, inquiry); /* start SCRIPT */
			break;

		  case 0x0a: /* read capacity command */
			if (!busy)
			 {

			  /* if stopped, we must still send back good status */
			  if (return_flag == 4)
				buffer_table[stat_buf].address = getPhysAddr(status_good);
			 }
			else

			  /* status must be good, even if busy */
			  dummy_table[stat_buf].address = getPhysAddr(status_good);
			write_long_reg(DSP, read_capacity); /* start SCRIPT */
			break;

		  case 0x0b: /* got a start/stop command */
			if (command_buf[4] & 1) /* This is a start command */
			 {

			  /* restore good status to buffer */
			  buffer_table[stat_buf].address = getPhysAddr(status_good);

			  /* clear the sense data buffer */
			  buffer_table[sense_data_buf].address=getPhysAddr(no_sense_data);
			  return_flag = 0; /* normal end of command */
			  siop_reg[SCRATCHA0] = 0x00; /* clear stopped flag */
			  siop_reg[SCRATCHA2] = 0x00; /* clear stopped/busy flag */
			  write_long_reg(DSP, tur); /* start SCRIPT */
			 }
			else /* this is a stop command */
			 {
			  return_flag = 1; /* at command complete, status to not ready */
			  siop_reg[SCRATCHA0] = 0x01; /* set stopped flag */
			  siop_reg[SCRATCHA2] = 0x01; /* set stopped/busy flag */
			  write_long_reg(DSP, stopped_busy_tur); /* start SCRIPT */
			 }
			break;

		  case 0x0c: /*  this is a format unit command */

			/* byte count is the entire size of the drive */
			byte_count = (unsigned long int)used_pages * page_size;

			/* seek to the first 4 logical pages */
			error = seek(0, 4, EMM_handle);

			/* k keeps track of the number of total bytes done, while j is the */
			/* current offset into the emm window */
			for(k = 0, j = 0, l = 4; k < byte_count ; k++, j++)
			 {

			  /* if we have finished the current page, then we need to seek to */
			  /* the next page */
			  if (j == (4 * page_size))
				{

				 /* seek to the next logical page */
				 error = seek(l, 4, EMM_handle);
				 if (error) /* error on seek is bad */
				  {
					error = release(EMM_handle);
					if (error)
					 printf("error deallocating pages\n");
					exit(1);
				  }
				 j = 0; /* return the index to zero */
				 l += 4; /* increment l to the next set of logical pages */
				}
			  medium_window[j] = 0x00; /* zero the current location */
			 }

			if (return_flag == 2)
			 {

			  /* if check condition exists, we must still send back good status */
			  buffer_table[stat_buf].address = getPhysAddr(status_good);

			  /* after command complete, return status to check condition */
			  return_flag=3;
			 }
			write_long_reg(DSP, tur); /* start SCRIPT */
			break;

		  case 0x0d: /* this is a send diagnostic command */
			if (command_buf[1] & 0x04) /* self test, return good status */
			 {
			  if (return_flag == 2)
				{

				 /* if check condition exists, we must still send good status */
				 buffer_table[stat_buf].address = getPhysAddr(status_good);

				 /* after command complete, return status to check condition */
				 return_flag=3;
				}
			  write_long_reg(DSP, tur); /* start SCRIPT */
			 }
			else /* this is an unsupported version of this command */
			 {

			  /* put check condition in status buffer */
			  buffer_table[stat_buf].address = getPhysAddr(check_condition);
			  return_flag = 2; /* at command complete, status back to good */

			  /* put illegal command sense data in the sense data buffer */
			  buffer_table[sense_data_buf].address = getPhysAddr(illegal_command);
			  write_long_reg(DSP, tur); /* start SCRIPT */
			 }
			break;

		  case 0x0e: /* received an abort message and disconnected */
			busy = 0; /* clear the driver busy flag */
			siop_reg[SCRATCHB1] = 0x00; /* clear the not done flag */
			siop_reg[SCRATCHA1] = 0x00; /* clear the SCRIPTS busy flag */
			write_long_reg(DSP, wait_select); /* start SCRIPT */
			if (return_flag != 4)
			 return_flag = 0; /* clear the return flag */
			break;

		  case 0x0f: /* this is an illegal command */

			/* put check condition in status buffer */
			buffer_table[stat_buf].address = getPhysAddr(check_condition);
			return_flag = 2; /* at command complete, status back to good */

			/* put illegal command sense data in the sense data buffer */
			buffer_table[sense_data_buf].address = getPhysAddr(illegal_command);
			write_long_reg(DSP, tur); /* start SCRIPT */
			break;

		  case 0x10: /* got SIGP and stopped */
			got_int = 1;
			break;

		  case 0x11: /* got selected while doing a reselection */

			/* put the dummy table address into the DSA register */
			write_long_reg(DSA, getPhysAddr(dummy_table));
			busy = 1; /* set the driver busy flag */
			siop_reg[SCRATCHA1] = 0x01; /* set the SCRIPTS busy flag */
			siop_reg[SCRATCHA2] = 0x01; /* set stopped/busy flag */
			write_long_reg(DSP, stopped_busy_wait_select); /* start SCRIPT */
			break;

		  case 0x12: /* done with command while busy */

			/* restore device busy to the status buffer in case it changed */
			dummy_table[stat_buf].address = getPhysAddr(device_busy);

			/* restore the regular table address to the DSA register */
			write_long_reg(DSA, getPhysAddr(buffer_table));
			busy = 0; /* clear the driver busy flag */
			siop_reg[SCRATCHA1] = 0x00; /* clear the SCRIPTS busy flag */
			siop_reg[SCRATCHA2] = 0x00; /* clear stopped/busy flag */
			switch(command_buf[0])
			 {
			  case 0x08: /* we were performing a 6-byte read */
				write_long_reg(DSP, read_reconnect); /* start SCRIPT */
				break;

			  case 0x0a: /* we were performing a 6-byte write */
				write_long_reg(DSP, write_reconnect); /* start SCRIPT */
				break;

			  case 0x28: /* we were performing a 10-byte read */
				write_long_reg(DSP, read_reconnect); /* start SCRIPT */
				break;

			  case 0x2a: /* we were performing a 10-byte write */
				write_long_reg(DSP, write_reconnect); /* start SCRIPT */
				break;
			 }
		 }
		break;

	  case 0x01: /* illegal instruction interrupt */
		printf("An Illegal Instruction Was Detected At SCRIPT Address ");

		/* print address of instruction that caused fault */
		for (i = 0; i < 4; i++)
		 addr[i] = siop_reg[DSP + i];
		printf("%02x%02x%02x%02xh\n", addr[3], addr[2], addr[1], addr[0]);

		/* print out the SCRIPTS base address */
		printf("SCRIPTS start at address %08lxh\n", scripts_address);
		printf("Note:  The DSP Points At The Next Instruction To Be Fetched\n");

		/* print out the instruction that caused the fault */
		dcmd = siop_reg[DCMD];
		for (i = 0; i < 4; i++)
		 dsps[i] = siop_reg[DSPS + i];
		for (i = 0; i < 3; i++)
		 dbc[i] = siop_reg[DBC + i];

		printf("DCMD = %02xh\n", dcmd);
		printf("DBC = %02x%02x%02xh\n", dbc[2], dbc[1], dbc[0]);
		printf("DSPS = %02x%02x%02x%02xh\n", dsps[3], dsps[2], dsps[1], dsps[0]);
		release_irq(); /* release interrupt and restore mask */
		error = release(EMM_handle);
		if (error)
		 printf("error deallocating pages\n");
		exit(1); /* terminate program with error code */
		break;

	  case 0x02: /* watchdog timeout interrupt */
		printf("A watchdog timeout was detected.\n");
		release_irq(); /* release interrupt and restore mask */
		error = release(EMM_handle);
		if (error)
		 printf("error deallocating pages\n");
		exit(1); /* terminate program with error code */
		break;

	  case 0x20: /* bus fault interrupt */
		printf("a bus fault occured\n");
		release_irq(); /* release interrupt and restore mask */
		error = release(EMM_handle);
		if (error)
		 printf("error deallocating pages\n");
		exit(1); /* terminate program with error code */
		break;
	 }
  }
 if (istat & 4) /* this is an interrupt on the fly that signifies end of i/o */
  {
	siop_reg[ISTAT] = 0x04; /* write back istat to clear the intfly */

	/* return flag determines the action to be taken at the end of the i/o */
	switch (return_flag)
	 {
	  case 0: /* normal end of command */
		break;

	  case 1: /* this is after a stop command terminates */

		/* put check condition in the status buffer */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);

		/* put unit not ready sense data in the sense data buffer */
		buffer_table[sense_data_buf].address = getPhysAddr(unit_not_ready);
		return_flag = 4; /* set the return flag for future commands */
		break;

	  case 2: /* this is after sending a check condition */

		/* after sending check condition, status goes back to good */
		buffer_table[stat_buf].address = getPhysAddr(status_good);
		return_flag = 0; /* clear the return flag */
		break;

	  case 3: /* this is after inquiry or format when check condition existed */

		/* put the check condition back in the status buffer */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);
		return_flag = 2; /* put the return flag back to the way it was */
		break;

	  case 4: /* any command while stopped */

		/* if a command that is legal while stopped is received, the status */
		/* sent back was good, and now we have to put the check condition */
		/* back into the status buffer and the sense data must return to what */
		/* it was before if it changed */
		buffer_table[stat_buf].address = getPhysAddr(check_condition);
		buffer_table[sense_data_buf].address = getPhysAddr(unit_not_ready);
		break;
	 }
  }
 if (istat & 2) /* this is a SCSI interrupt */
  {

	/* read the two SCSI interrupt registers */
	sist0 = siop_reg[SIST0];
	sist1 = siop_reg[SIST1];
	if (sist0 & 0x80) /* ATN is active on the SCSI bus */
	 {
	  dsp = ((((uquad)siop_reg[DSP + 3]) << 24) /* get the current DSP value */
			 | (((uquad)siop_reg[DSP + 2]) << 16)
			 | (((uquad)siop_reg[DSP + 1]) << 8)
			 | siop_reg[DSP]);

	  /* subtract 8 to point to the instruction that caused the initiator */
	  /* to assert ATN */
	  dsp -= 8;

	  /* go to message out phase to see what the inititor wants */
	  write_long_reg(DSP, msg_out_phase);
	 }
	if (sist0 & 1) /* SCSI parity error */
	 {
	  phase = (siop_reg[DCMD] & 0x07); /* check the current phase */

	  /* if this was a message parity error, we must reject it */
	  if (phase == 0x06)
		{

		 /* set up the message in the table */
		 buffer_table[msg_in_buf].address = getPhysAddr(message_reject);
		 return_code = 1; /* says we sent message reject */
		}
	  else /* for any other phase, send a restore pointers and try again */
		{

		 /* set up the message in the table */
		 buffer_table[msg_in_buf].address = getPhysAddr(restore_pointers_msg);

		 /* says we got a parity error, we should start on the last instruction */
		 return_code = 2;
		 dsp = ((((uquad)siop_reg[DSP + 3]) << 24) /* get the current DSP value */
				| (((uquad)siop_reg[DSP + 2]) << 16)
				| (((uquad)siop_reg[DSP + 1]) << 8)
				| siop_reg[DSP]);

		 /* subtract 8 to point to the instruction that was running when the */
		 /* parity error was received */
		 dsp -= 8;
		}

	  /* go to message in phase and send our message */
	  write_long_reg(DSP, msg_in_phase);
	 }
	if (sist0 & 2) /* a SCSI bus reset has occured */
	 {

	  /* call the function to reset the chip and set up initial values */
	  init_siop();

	  /* write the buffer table address into the DSA register */
	  write_long_reg(DSA, getPhysAddr(buffer_table));

	  /* put the check condition in the status buffer */
	  buffer_table[stat_buf].address = getPhysAddr(check_condition);

	  /* put command complete into the message in buffer */
	  buffer_table[msg_in_buf].address = getPhysAddr(command_complete);
	  return_flag = 2; /* we sent a check condition */

	  /* put the unit attention sense data in the sense data buffer */
	  buffer_table[sense_data_buf].address = getPhysAddr(reset_sense_data);
	  write_long_reg(DSP, wait_select); /* start SCRIPT */

	  /* return needed variables to their initial values */
	  first = 1;
	  done_synch = 0;
	  done_wide = 0;
	 }
  }

 /* write the two end of interrupt registers */
 outportb(PIC2_EOI_REG, 0x20);
 outportb(PIC1_EOI_REG, 0x20);

 enable(); /* enable system interrupts */
}

/* init_dma - enables DMA channel 5 for use by 16-bit master */
static void
init_dma()
{
	 outportb(0x00D2, 0x01); /* clear "manual" REQ bit */
	 outportb(0x00D4, 0x05); /* mask REQ */
	 outportb(0x00D8, 0x00); /* clear byte pointer flip-flop */
	 outportb(0x00C6, 0x00); /* transfer length: allow one transfer */
	 outportb(0x00C6, 0x00);
	 outportb(0x00D6, 0xD9); /* READ, cascade mode, autoinit */
	 outportb(0x00D4, 0x01); /* unmask REQ */
}

static void
init_siop()
{
 uword respid;

 /* Set the pointer siop_reg to point to the base of the c720's registers. */
 /* This allows direct and simple access to the chips registers. */
 siop_reg = MK_FP(SIOP_SEGMENT_ADDRESS, SIOP_OFFSET_ADDRESS);

 siop_reg[ISTAT] |= 0x40; /* reset 720 */
 siop_reg[ISTAT] ^= 0x40;

 siop_reg[SCNTL0] = 0xcd; /* set target mode, parity checking and generation */
 siop_reg[SCNTL1] |= 0x20; /* set DHP */

 /* set ID and enable response to selection */
 siop_reg[SCID] = (ubyte)(0x20 | (id & 0x0f));

 respid = 1; /* put a 1 in the variable */

 /* this moves the bit to the spot that corresponds to the chosen id */
 respid = (respid << id);
 siop_reg[RESPID0] = (ubyte)(respid & 0xff); /* set response ID */
 siop_reg[RESPID1] = (ubyte)((respid >> 8) & 0xff);

 /* set up clock dividers based on the SCLK value provided by user */
 if ((clk > 16) && (clk <= 25))
  siop_reg[SCNTL3] = 0x11;
 else
  if ((clk > 25) && (clk <= 38))
	siop_reg[SCNTL3] = 0x22;
  else
	if ((clk > 38) && (clk <= 50))
	 siop_reg[SCNTL3] = 0x33;
	else
	 if ((clk > 50) && (clk <= 66))
	  siop_reg[SCNTL3] = 0x44;
	 else
	  if ((clk > 66) && (clk <= 80))
		siop_reg[SCNTL3] = 0x55;

 /* set up the reselection table entry */
 sel_info[3] = siop_reg[SCNTL3];
 sel_info[1] = 0;
 buffer_table[selector_id].count = (((unsigned long int)sel_info[3] << 24)
											 | ((unsigned long int)sel_info[2] << 16)
											 | ((unsigned long int)sel_info[1] << 8)
											 | sel_info[0]);
 siop_reg[SXFER] = 0x00; /* set asynchronous */
 siop_reg[STEST3] = 0x80; /* enable tolerant */

 /* set up the SCSI and DMA interrupt masks */
 siop_reg[DIEN] = 0x77;
 siop_reg[SIEN0] = 0x8f;

 siop_reg[CTEST0] = 0x80; /* Cache Burst Disable */
 siop_reg[DMODE] = 0x80; /* 8 Transfer Burst */
}

static void
init_table()
{

 /* Allocate space for the table of addresses used by the table indirect */
 /* commands.  This table must be on a 32-bit boundary. */
 buffer_table = (table *) my_malloc((10+TABLE_SIZE) * sizeof(table));
 if (FP_OFF(buffer_table) & 0x03)
  {
	printf("buffer_table not word aligned.  Fix program.\n");
	error = release(EMM_handle);
	if (error)
	 printf("error deallocating pages\n");
	exit(1);
  }

 /* put the table address into the DSA register */
 write_long_reg(DSA, getPhysAddr(buffer_table));

 /* create the dummy table for use while busy */
 dummy_table = (table *) my_malloc((10+TABLE_SIZE) * sizeof(table));
 if (FP_OFF(dummy_table) & 0x03)
  {
	printf("dummy_table not word aligned.  Fix program.\n");
	error = release(EMM_handle);
	if (error)
	 printf("error deallocating pages\n");
	exit(1);
  }
}

static void
init_script()
{
 /* Make sure script starts on a 32-bit boundary */
 if (FP_OFF(SCRIPT) & 0x03)
  {
	script = (uquad *) my_malloc(sizeof(SCRIPT));
	if (FP_OFF(script) & 0x03)
	 {
	  printf("script not word aligned.  Fix program.\n");
	  error = release(EMM_handle);
	  if (error)
		printf("error deallocating pages\n");
	  exit(1);
	 }
	memcpy(script, SCRIPT, sizeof(SCRIPT));
  }
 else
  script = SCRIPT;
}

static void
init_irq()
{
 disable(); /* disable interrupts */

 if (irq == 11)
  {
	setvect(IRQ11, isr); /* set irq interrupt vector */
	old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */
	outportb(PIC2_MASK_REG, IRQ11_mask & old_irq_mask); /* enable irq mask */
  }
 if (irq == 12)
  {
	setvect(IRQ12, isr); /* set irq interrupt vector */
	old_irq_mask = inportb(PIC2_MASK_REG); /* get old int mask */
	outportb(PIC2_MASK_REG, IRQ12_mask & old_irq_mask); /* enable irq mask */
  }
 enable(); /* enable interrupts */
}

static void
release_irq()
{
 disable();	/* disable interrupts */
 outportb(PIC2_MASK_REG, old_irq_mask); /* restore the old mask */
 enable(); /* enable interrupts */
}

